//-----------------------------------------------------------------------------
/*
Programmed by Olivier Sangala
http://olivier.sangala.free.fr
olivier_sangala@hotmail.com
*/
//-----------------------------------------------------------------------------

#include <tigcclib.h>

//-----------------------------------------------------------------------------

#define MAP_W (7)
#define MAP_H (7)

//-----------------------------------------------------------------------------

enum {WAY_L,WAY_R,WAY_U,WAY_D};

//-----------------------------------------------------------------------------

typedef struct
{
	int num;
	int used;
	int pathrank;
}
TTile;

typedef struct
{
	int x,y;
	int n;
	int way;
}
TPos;

typedef struct
{
	TPos P[MAP_W*MAP_H];
	int Count;
}
TPList;

//-----------------------------------------------------------------------------

void Init(void);
void MakeS1(void);
void PL_AddAllDests(int x, int y);
void PL_AddDests(int x, int y, int n);
int IsCorrectPos(int x, int y);
int IsTileS1(int x, int y);
int IsTileS2(int x, int y);
void MakeS2(void);
int S2_GetScore(void);
void MakeAll(void);
void MoveCursor(int way);

void DrawScreen(void);


//-----------------------------------------------------------------------------

TTile Tiles[MAP_H][MAP_W];
TPList *PL;

int Mode12;
int PathLength;

int px,py;
int sx=0,sy=0;
int dx,dy;

//-----------------------------------------------------------------------------

unsigned short TileDataS[16]=
{
0xFFC0,0xFFC0,0xFFC0,0xFFC0,0xFFC0,0xFFC0,0xFFC0,0xFFC0,
0xFFC0,0xFFC0,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
};

unsigned short TileDataD[16]=
{
0xFFC0,0x8040,0x8040,0x8040,0x8040,0x8040,0x8040,0x8040,
0x8040,0xFFC0,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
};


//-----------------------------------------------------------------------------

void Init(void)
{
	int x,y;
	for (y=0;y<MAP_H;y++)
	for (x=0;x<MAP_W;x++)
	{
		Tiles[y][x].num=-1;
		Tiles[y][x].used=0;
		Tiles[y][x].pathrank=-1;
	}
}
//-----------------------------------------------------------------------------

void MakeS1(void)
{
	int cx=sx;
	int cy=sy;
	
	PathLength=0;
	Mode12=1;
	
	for (;;)
	{
		PL->Count=0;
		PL_AddAllDests(cx,cy);
		
		//printf("Step \n");ngetchx();
		if (PL->Count==0)
		break;
		
		int r=random(PL->Count);
		int n=PL->P[r].n;
		
		Tiles[cy][cx].num=n;
		//Tiles[cy][cx].used=1;
		Tiles[cy][cx].pathrank=(PathLength++);
		
		int a;
		for (a=0;a<PL->Count;a++)
		if (PL->P[a].n==n)
		Tiles[PL->P[a].y][PL->P[a].x].used=1;
		
		cx=PL->P[r].x;
		cy=PL->P[r].y;
		
	}
	
	dx=cx;
	dy=cy;
	
	Tiles[dy][dx].num=0;
	Tiles[dy][dx].pathrank=(PathLength++);
	
}
//-----------------------------------------------------------------------------


void PL_AddAllDests(int x, int y)
{
	int n;
	for (n=1;n<=(max(MAP_W,MAP_H)-1);n++) // DEBUG (n<=6)
	PL_AddDests(x,y,n);
}
//-----------------------------------------------------------------------------


void PL_AddDests(int x, int y, int n)
{
	int xx,yy;
	
	int way;
	for (way=0;way<4;way++)
	{
		xx=x;
		yy=y;
		
		switch (way)
		{
			case WAY_L : xx-=n; break;
			case WAY_R : xx+=n; break;
			case WAY_U : yy-=n; break;
			case WAY_D : yy+=n; break;
		}
		
		if (IsCorrectPos(xx,yy))
		if ((Mode12==1 && IsTileS1(xx,yy)) || (Mode12==2 && IsTileS2(xx,yy)))
		{
			// add pos
			PL->P[PL->Count].x=xx;
			PL->P[PL->Count].y=yy;
			PL->P[PL->Count].n=n;
			PL->P[PL->Count].way=way;
			PL->Count++;
		}
	}
}
//-----------------------------------------------------------------------------

int IsCorrectPos(int x, int y)
{
	return (x>=0 && x<MAP_W && y>=0 && y<MAP_H);
}
//-----------------------------------------------------------------------------

int IsTileS1(int x, int y)
{
	return (Tiles[y][x].num==-1 && Tiles[y][x].used==0);
}
//-----------------------------------------------------------------------------

int IsTileS2(int x, int y)
{
	return (Tiles[y][x].num!=-1);
}
//-----------------------------------------------------------------------------

void MakeS2(void)
{
	Mode12=2;
	
	int x,y;
	for (y=0;y<MAP_H;y++)
	for (x=0;x<MAP_W;x++)
	if (!(x==dx && y==dy))
	if (Tiles[y][x].num==-1)
	{
		int nb=1;
		int sb=MAP_W*MAP_H;
		
		int n;
		for (n=1;n<=(max(MAP_W,MAP_H)-1);n++) // DEBUG (n<=6)
		{
			PL->Count=0;
			PL_AddDests(x,y,n);
			int s=S2_GetScore();
			
			if (s<sb)
			{
				sb=s;
				nb=n;
			}
		}
		
		Tiles[y][x].num=nb;
		Tiles[y][x].pathrank=sb;
	}
	
}
//-----------------------------------------------------------------------------

int S2_GetScore(void)
{
	int sb=0;
	
	int a;
	for (a=0;a<PL->Count;a++)
	{
		if (Tiles[PL->P[a].y][PL->P[a].x].pathrank>sb)
		sb=Tiles[PL->P[a].y][PL->P[a].x].pathrank;
	}
	
	return sb;
}
//-----------------------------------------------------------------------------

void MakeAll(void)
{
	
	Init();
	MakeS1();
	MakeS2();
	
	px=sx;
	py=sy;
	
}
//-----------------------------------------------------------------------------

void MoveCursor(int way)
{
	
	int n=Tiles[py][px].num;
	int xx=px;
	int yy=py;
	
	switch (way)
	{
		case WAY_L : xx-=n; break;
		case WAY_R : xx+=n; break;
		case WAY_U : yy-=n; break;
		case WAY_D : yy+=n; break;
	}
	
	if (IsCorrectPos(xx,yy))
	{
		px=xx;
		py=yy;
	}
	
}
//-----------------------------------------------------------------------------

void DrawScreen(void)
{
	ClrScr();
	
	int x,y;
	for (y=0;y<MAP_H;y++)
	for (x=0;x<MAP_W;x++)
	{
		char ch='0'+Tiles[y][x].num;
		DrawChar(10*x+2,10*y+1,ch,A_NORMAL);
	}
	
	
	Sprite16(10*dx,10*dy,10,TileDataD,LCD_MEM,SPRT_XOR);
}
//-----------------------------------------------------------------------------

// Main Function
void _main(void)
{
	TPList PLA;
	PL=&PLA;
	
	randomize();
	
	clrscr();
	FontSetSys(0);
	printf("RideMaze 1.0 \n");
	printf("Programmed by Olivier Sangala \n");
	printf("http://olivier.sangala.free.fr \n");
	printf("olivier_sangala@hotmail.com \n");
	FontSetSys(1);
	ngetchx();
	
	MakeAll();
	DrawScreen();
	
	for (;;)
	{
		
		if (px==dx && py==dy)
		DrawStr(100,0,"You Won!",A_XOR);
		
		Sprite16(10*px,10*py,10,TileDataS,LCD_MEM,SPRT_XOR);
		int key=ngetchx();
		Sprite16(10*px,10*py,10,TileDataS,LCD_MEM,SPRT_XOR);
		
		if (px==dx && py==dy)
		DrawStr(100,0,"You Won!",A_XOR);
		
		if (key==KEY_LEFT)
		MoveCursor(WAY_L);
		else if (key==KEY_RIGHT)
		MoveCursor(WAY_R);
		else if (key==KEY_UP)
		MoveCursor(WAY_U);
		else if (key==KEY_DOWN)
		MoveCursor(WAY_D);
		
		else if (key==KEY_F2)
		{
			MakeAll();
			DrawScreen();
		}
		else if (key==KEY_CLEAR)
		{
			px=sx;
			py=sy;
		}
		
		else if (key==KEY_F3)
		{
			char buffer[16];
			sprintf(buffer,"%d",PathLength);
			DrawStr(100,10,buffer,A_NORMAL);
			ngetchx();
			DrawScreen();
		}
		
		else if (key==KEY_APPS)
		off();
		else if (key==KEY_ESC)
		break;
	}
	
}
